import gc
import sys
import ast
import inspect
from string import Formatter

from f.chain_unit import ChainUnit
from f.lazy_string import LazyString


class ProxyModule(sys.modules[__name__].__class__):
    old_str = str

    def __call__(self, string, lazy=True, safe=True, closures=True):
        if isinstance(string, LazyString):
            return string

        base_frame = inspect.stack(0)[1].frame

        result = LazyString(
            [ChainUnit(base=x[0], appendix=x[1], lazy=lazy) for x in Formatter().parse(string)],
            {**base_frame.f_locals},
            {**base_frame.f_globals},
            self.sum_of_nonlocals(
                base_frame.f_back,
                self.get_qualname(base_frame.f_code, raise_if_not_literal=safe, code_line=base_frame.f_lineno),
                closures,
                safe,
            ),
            lazy,
        )

        if lazy:
            return result
        return result.data

    def sum_of_nonlocals(self, first_frame, base_qualname, closures, safe):
        if not closures or first_frame is None or base_qualname is None:
            return {}

        all_locals = []
        while first_frame is not None:
            code = first_frame.f_code

            qualname = self.get_qualname(code, raise_if_not_literal=False, code_line=0)
            if qualname is not None:
                if self.startswith(base_qualname.split('.'), qualname.split('.')):
                    all_locals.append(first_frame.f_locals)

            first_frame = first_frame.f_back

        result = {}
        index = len(all_locals) - 1

        while index >= 0:
            result.update(all_locals[index])
            index -= 1

        return result

    @classmethod
    def get_qualname(cls, code, raise_if_not_literal, code_line):
        functions = []

        for function in gc.get_referrers(code):
            maybe_code = None
            if inspect.isgenerator(function):
                maybe_code = getattr(function, 'gi_code', None)
            elif callable(function):
                maybe_code = getattr(function, '__code__', None)
            if maybe_code is not None:
                functions.append(function)

        if functions:
            function = functions[0]
            if raise_if_not_literal:
                cls.check_code(function, code, code_line)
            return function.__qualname__

    @staticmethod
    def check_code(function, code, code_line):
        try:
            if inspect.isgenerator(function):
                code_strings, begin_code_line_number = inspect.getsourcelines(code)
            else:
                code_strings, begin_code_line_number = inspect.getsourcelines(function)
        except:
            return

        spaces_count = 0
        for letter in code_strings[0]:
            if not letter.isspace():
                break
            spaces_count += 1

        code_strings = [x[spaces_count:] for x in code_strings]

        full_code = ''.join(code_strings)
        ast_of_code = ast.parse(full_code)
        expected_line_number = begin_code_line_number

        flag = True
        class ConstantVisitor(ast.NodeVisitor):
            def visit_Call(self, node):
                nonlocal flag
                if node.lineno + begin_code_line_number - 1 == code_line:
                    if hasattr(node.func, 'id') and node.func.id == 'f':
                        if len(node.args) == 1 and isinstance(node.args[0], ast.Constant):
                            flag *= True
                        elif len(node.args) == 1 and isinstance(node.args[0], ast.Call) and hasattr(node.args[0].func, 'id') and node.args[0].func.id == 'f':
                            flag *= True
                            for arg in node.args:
                                ConstantVisitor().visit(arg)
                            for keyword in node.keywords:
                                ConstantVisitor().visit(keyword.value)
                        else:
                            flag *= False
                    elif isinstance(node.func, ast.Attribute):
                        ConstantVisitor().visit(node.func.value)
                        for arg in node.args:
                            ConstantVisitor().visit(arg)
                        for keyword in node.keywords:
                            ConstantVisitor().visit(keyword.value)
                    else:
                        for arg in node.args:
                            ConstantVisitor().visit(arg)
                        for keyword in node.keywords:
                            ConstantVisitor().visit(keyword.value)

        ConstantVisitor().visit(ast_of_code)

        if not flag:
            raise SyntaxError('Unsafe use of a variable as a template.')


    @staticmethod
    def startswith(iterable, second_iterable):
        if len(iterable) < len(second_iterable):
            return False

        for element, second_element in zip(second_iterable, iterable):
            if element != second_element:
                return False

        return True

    def __str__(self):
        return 'f'

    def __repr__(self):
        return 'f'



class ChainUnit:
    def __init__(self, base, appendix=None, lazy=True):
        self.base = base
        self.appendix = appendix
        if self.appendix is not None:
            if self.appendix == '':
                raise SyntaxError('lazy f-string: empty expression not allowed')



from collections import UserString


class LazyString(UserString, str):
    __slots__ = ('units', 'local_locals', 'local_globals', 'local_nonlocals', 'lazy', 'result')

    def __init__(self, units, local_locals, local_globals, local_nonlocals, lazy):
        self.units = units
        self.local_locals = local_locals
        self.local_globals = local_globals
        self.local_nonlocals = local_nonlocals
        self.lazy = lazy
        self.result = None

    def __new__(cls, *args, **kwargs):
        return str.__new__(cls)

    def __add__(self, other):
        if isinstance(other, type(self)):
            other = other.data
        return self.data.__add__(other)

    def __radd__(self, other):
        return other + self.data

    def __getitem__(self, index):
        return self.data.__getitem__(index)

    def __getnewargs__(self):
        return (self.units, self.local_locals, self.local_globals, self.local_nonlocals, self.lazy)

    def __mod__(self, args):
        if isinstance(args, type(self)):
            args = args.data
        return self.data.__mod__(args)

    def __rmod__(self, template):
        return str(template) % self.data

    def __mul__(self, n):
        return self.data * n

    def __rmul__(self, n):
        return self.data * n

    def __ne__(self, other):
        if isinstance(other, type(self)):
            other = other.data
        return self.data != other

    def __reduce__(self):
        raise TypeError('cannot pickle {0} object'.format(type(self).__name__))

    def __setattr__(self, name, value):
        if name not in type(self).__slots__:
            raise AttributeError(
                "'{0}' object has no attribute '{1}'".format(type(self).__name__, name)
            )
        object.__setattr__(self, name, value)

    def replace(self, old, new, maxsplit=-1):
        if isinstance(old, type(self)):
            old = old.data
        if isinstance(new, type(self)):
            new = new.data
        return self.data.replace(old, new, maxsplit)

    def split(self, sep=None, maxsplit=-1):
        if isinstance(sep, type(self)):
            sep = sep.data
        return self.data.split(sep, maxsplit)

    def rsplit(self, sep=None, maxsplit=-1):
        if isinstance(sep, type(self)):
            sep = sep.data
        return self.data.rsplit(sep, maxsplit)

    def lower(self):
        return self.data.lower()

    def upper(self):
        return self.data.upper()

    def zfill(self, width):
        return self.data.zfill(width)

    def translate(self, *args):
        return self.data.translate(*args)

    def title(self):
        return self.data.title()

    def startswith(self, prefix, *other_args):
        if isinstance(prefix, type(self)):
            prefix = prefix.data
        return self.data.startswith(prefix, *other_args)

    def endswith(self, suffix, *other_args):
        if isinstance(suffix, type(self)):
            suffix = suffix.data
        return self.data.endswith(suffix, *other_args)

    def index(self, sub, *other):
        if isinstance(sub, type(self)):
            sub = sub.data
        return self.data.index(sub, *other)

    def rindex(self, sub, *other):
        if isinstance(sub, type(self)):
            sub = sub.data
        return self.data.rindex(sub, *other)

    def center(self, width, *others):
        others = [other.data if isinstance(other, type(self)) else other for other in others]
        return self.data.center(width, *others)

    def join(self, iterable):
        iterable = [item.data if isinstance(item, type(self)) else item for item in iterable]
        return self.data.join(iterable)

    def capitalize(self):
        return self.data.capitalize()

    def encode(self, **kwargs):
        return self.data.encode(**kwargs)

    def casefold(self):
        return self.data.casefold()

    def expandtabs(self, tabsize=8):
        return self.data.expandtabs(tabsize)

    def removeprefix(self, prefix):
        if isinstance(prefix, type(self)):
            prefix = prefix.data
        return self.data.removeprefix(prefix)

    def removesuffix(self, suffix):
        if isinstance(suffix, type(self)):
            suffix = suffix.data
        return self.data.removesuffix(suffix)

    def lstrip(self, chars=None):
        if isinstance(chars, type(self)):
            chars = chars.data
        return self.data.lstrip(chars)

    def rstrip(self, chars=None):
        if isinstance(chars, type(self)):
            chars = chars.data
        return self.data.rstrip(chars)

    def ljust(self, width, *args):
        args = [item.data if isinstance(item, type(self)) else item for item in args]
        return self.data.ljust(width, *args)

    def rjust(self, width, *args):
        args = [item.data if isinstance(item, type(self)) else item for item in args]
        return self.data.rjust(width, *args)

    def encode(self, **kwargs):
        kwargs = {key: value.data if isinstance(value, type(self)) else value for key, value in kwargs.items()}
        return self.data.encode(**kwargs)

    @staticmethod
    def maketrans(x, *others):
        others = [item.data if isinstance(item, UserString) else item for item in others]
        x = x.data if isinstance(x, UserString) else x

        if isinstance(x, str):
            return str.maketrans(x, *others)

        first_item = {}
        for key, value in x.items():
            key = key.data if isinstance(key, UserString) else key
            value = value.data if isinstance(value, UserString) else value
            first_item[key] = value

        return str.maketrans(first_item, *others)

    def partition(self, sep):
        sep = sep.data if isinstance(sep, type(self)) else sep
        return self.data.partition(sep)

    def rpartition(self, sep):
        sep = sep.data if isinstance(sep, type(self)) else sep
        return self.data.rpartition(sep)

    def swapcase(self):
        return self.data.swapcase()

    @property
    def data(self):
        if self.result is not None:
            return self.result

        result = []

        for unit in self.units:
            result.append(unit.base)
            if unit.appendix is not None:
                substring = 'x = {0}'.format(unit.appendix)
                namespace = self.local_globals.copy()
                namespace.update(self.local_nonlocals)
                namespace.update(self.local_locals)
                try:
                    exec(substring, namespace)
                except SyntaxError as e:
                    raise SyntaxError('lazy f-string: invalid syntax') from e
                result.append(str(namespace['x']))

        self.result = ''.join(result)
        del self.units
        del self.local_locals
        del self.local_nonlocals
        del self.local_globals

        return self.result



from setuptools import setup, find_packages

with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

requirements = []

setup(
    name='fazy',
    version='0.0.4',
    author='Evgeniy Blinov',
    author_email='zheni-b@yandex.ru',
    description='Lazy f-strings for everyone',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/pomponchik/fazy',
    packages=find_packages(exclude=('tests',)),
    install_requires=requirements,
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
    ],
)
